public class Model {
	Point center;
	PointList points;
	PolyList polygons;
	boolean abs = false;
	double bounding_sphere;
	//-----------------------------------------
	//constructor
	//-----------------------------------------
	public Model() {
		this.center = new Point();
		this.points = new PointList();
		this.polygons = new PolyList();
	}
	public Model(int numpoint, int numpoly) {
		this.center = new Point();
		this.points = new PointList(numpoint);
		this.polygons = new PolyList(numpoly);
	}
	public Model(String str) {
		this.center = new Point();
		this.points = new PointList();
		this.polygons = new PolyList();
		this.compose(str);
	}
	public Model(String str, int numpoint, int numpoly) {
		this.center = new Point();
		this.points = new PointList(numpoint);
		this.polygons = new PolyList(numpoly);
		this.compose(str);
	}
	//----------------------------------------
	//ultimately usefull
	//----------------------------------------
	//all this because i wanted point to never duplicate , but i wanted to be able
	//to duplicate an entire model , and the easiest way was to keep only index to
	//point in polygon . so polygon keep a number who index a point in the point list.
	//then just give the point number in polygon and the polygon number , it will return the point
	//this.points.list = { Point1, Point2, ...}
	//this.polygons.list = { Poly1, Poly2, ...}
	//this.polygons.list[0] = Poly1;
	//Poly1.list[0] = 1 (example that the first point of the poly is the Point2);
	//this.points.list[1] = Point2;
	public Point get_point_of_poly(int numpoint, int numpoly) {
		return this.points.list[ this.polygons.list[numpoly].list[numpoint] ];
	}
	//----------------------------------------
	//collision detection
	//----------------------------------------
	//the perfect collision detection for a space asteroid game
	public boolean collide_with_bounding_sphere_of(Point center, double bounding_sphere) {
		//the first step for collision detection is the bounding sphere
		double dist = new Point(this.center).dist(center);
		if (dist < this.bounding_sphere+bounding_sphere) {
			return true;
		}
		return false;
	}
	public boolean collide_with_bounding_sphere_of(Model model) {
		return collide_with_bounding_sphere_of(model.center, model.bounding_sphere);
	}
	//calcutation the normal of three point (p1,p2,p3), is making the cross product of two line (p1,p2 and p2,p3)
	//normal = cross_product((p1-p2),(p3-p2))
	//k = dot_product(p1,normal);
	//to know if the point is in plane;
	//dot_product(point , normal) = k           Point is in plane 
	//dot_product(point , normal) > k           Point is on one side of plane 
	//dot_product(point , normal) < k           Point is on other side of plane
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	public Point normal_of_poly(int polynum) {
		Point p1 = this.get_point_of_poly(0, polynum);
		Point p2 = this.get_point_of_poly(1, polynum);
		Point p3 = this.get_point_of_poly(2, polynum);
		//return new Point(p1).sub(p2).cross_product(new Point(p3).sub(p2));	//too much new Point , evitate too much allocation
		//make it the hard way
			double ax = p1.x-p2.x, ay = p1.y-p2.y, az = p1.z-p2.z;
			double bx = p3.x-p2.x, by = p3.y-p2.y, bz = p3.z-p2.z;
			return new Point( (ay*bz)-(az*by), (az*bx)-(ax*bz), (ax*by)-(ay*bx) );	//cross product
	}
	//the perfect collison detection for wall
	public boolean planes_vs_bounding_sphere_of(Point point, double bounding_sphere) {
		this.absolute();
		int cnt = 0; 
		for (int i = 0; i < this.polygons.num; i++) {
			Point normal = this.normal_of_poly(i);
			double dist = normal.dot_product( new Point( point ).sub(this.center) );
			if (dist < bounding_sphere && dist > -bounding_sphere) {
				cnt += 1;
			}
		}
		if (cnt >= this.polygons.num) {
			return true;
		}
		return false;
	}
	public boolean planes_vs_bounding_sphere_of(Model model) {
		return planes_vs_bounding_sphere_of(model.center, model.bounding_sphere);
	}

	//the direction and the normal have to be normalised (enormous overhead)
	//this function compare with the plane and not the polygon ...
	//i just dont know what to do with it
	public Point collision_point_of_a_direction_and_a_plane(Point p1, Point normal, Point center, Point direction) {
		double time = -( (normal.dot_product(center) + normal.dot_product(p1)) / normal.dot_product(direction) );
		return new Point(center).add(new Point(direction).mul(time)); 
	}
	//----------------------------------------
	//opperation transmitted to point
	//----------------------------------------
	public void move(double x, double y, double z){
		if (this.abs == true) {
			for (int i = 0; i<this.points.num; i++) {
				this.points.list[i].add(x,y,z);
			}
		}
		center.add(x,y,z);
	}
	public void rotate_x(double r) {
		for (int i = 0; i<this.points.num; i++) {
			this.points.list[i].rotate_x(r);
		}
	}
	public void rotate_y(double r) {
		for (int i = 0; i<this.points.num; i++) {
			this.points.list[i].rotate_y(r);
		}
	}
	public void rotate_z(double r) {
		for (int i = 0; i<this.points.num; i++) {
			this.points.list[i].rotate_z(r);
		}
	}
	//-------------------------------------------
	//absolute/relative  if relative , all point are around center
	//-------------------------------------------
	public void absolute() {
		//point + vector
		if (this.abs == false) {
			for (int i = 0; i<this.points.num; i++) {
				this.points.list[i].add(this.center);
			}
			this.abs = true;
		}
	}
	public void relative() {
		//point - vector
		if (this.abs == true) {
			for (int i = 0; i<this.points.num; i++) {
				this.points.list[i].sub(this.center);
			}
			this.abs = false;
		}
	}
	//-------------------------------------------
	//compositor
	//-------------------------------------------
		//do something about optional backface culling
	public void compose(String str) {
			this.bounding_sphere = 0;		//determine the bounding sphere with the point who is the fartest from the center
			String[] polys = str.split("\n");				//extract polygon from string (separated by newline)
			for (int i=0; i<polys.length; i++) {
				String[] points = polys[i].split(" ");		//extract point from polygon string (sparated by whitespace)
				Poly poly = new Poly(points.length);
				for (int j = 0; j<points.length; j++) {
					String[] comps = points[j].split(":");	//extract x,y,z from point string (separated by : like x:y:z) 
					double x = Double.parseDouble(comps[0]);
					double y = Double.parseDouble(comps[1]);
					double z = Double.parseDouble(comps[2]);
					//if it is the first trio , then it is the color
					if (j == 0) {
						poly.fill((int)(x*255),(int)(y*255),(int)(z*255));
					}
					else if (j == 1) {
						poly.draw((int)(x*255),(int)(y*255),(int)(z*255));
					}
					else {
						//add point only if not already in list
						int point = -1;
						for (int h = 0; h < this.points.num; h++) {
							if ( (this.points.list[h].x == x)&&(this.points.list[h].y == y)&&(this.points.list[h].z == z) ) {
								point = h;
							}
						}//because many polygons can share the same point
						if (point == -1) {
							point = this.points.append(new Point(x,y,z));	//return point number in the points.list[]
							double tmpz = (x*x)+(y*y)+(z*z);	//bounding_sphere
							if (tmpz > this.bounding_sphere) {
								this.bounding_sphere = tmpz;
							}
						}
						//the second argument is not to plug but the face vector
						//still on the point list, but un connected with other in polygon
						//polygon keep it in face argument, use it for back face culling
						if (point != -1) {
							poly.append(point);		//need point number as argument
						}
					}
				}
				this.polygons.append(poly);
			}
	}
	//-------------------------------------------------------------------------------------------------------------------------------------------------
	//end
	//-------------------------------------------------------------------------------------------------------------------------------------------------
}
